#include "../include/dynbuf.h"

const int dynstreambuf::_blksize = 1024;

dynstreambuf::dynstreambuf() : streambuf(), buf_(0), bufc_(0), alloc_(0) {
}
dynstreambuf::~dynstreambuf() {
	int i;
	if(alloc_) {
		for(i = 0; i < bufc_; i++) {
			if(buf_[i])
				delete[] buf_[i];
			delete[] buf_;
		}
	}
}
int dynstreambuf::doallocate() {
	setb((char*)1, (char*)1);
	buf_ = new char*[1];
	buf_[0] = new char[_blksize];
	bufc_ = 1;
	setg(buf_[0], buf_[0], buf_[0]);
	setp(buf_[0], buf_[0] + _blksize);
	alloc_ = 1;
	return 1;
}
int dynstreambuf::sync() {
	return (gptr() == pptr() ? 0 : EOF);
}
int dynstreambuf::underflow() {
	char** tbuf;
	int i;
	if(!alloc_)
		return EOF;
	if(egptr() != buf_[0] + _blksize) {
		if(bufc_ == 1) {
			if(gptr() != pptr()) {
				setg(eback(), gptr(), pptr());
				return sgetc();
			}
		}
		else {
			setg(eback(), gptr(), buf_[0] + _blksize);
			return sgetc();
		}
	}
	else if(bufc_ > 1) {
		tbuf = new char*[--bufc_];
		for(i = 0; i < bufc_; i++)
			tbuf[i] = buf_[i + 1];
		delete[] buf_[0];
		delete[] buf_;
		buf_ = tbuf;
		setg(buf_[0], buf_[0], (bufc_ > 1 ? buf_[0] + _blksize : pptr()));
		return sgetc();
	}
	return EOF;
}
int dynstreambuf::overflow(int _c) {
	char** tbuf;
	int i;
	if(!alloc_)
		return (allocate() != EOF && _c != EOF ? sputc(_c) : EOF);
	if(pptr() >= epptr()) {
		tbuf = new char*[++bufc_];
		for(i = 0; i < bufc_ - 1; i++)
			tbuf[i] = buf_[i];
		tbuf[bufc_ - 1] = new char[_blksize];
		delete[] buf_;
		buf_ = tbuf;
		setp(buf_[bufc_ - 1], buf_[bufc_ - 1] + _blksize);
	}
	return (_c != EOF ? sputc(_c) : EOF);
}
int dynstreambuf::pbackfail(int _c) {
	char** tbuf;
	int i;
	if(!alloc_)
		if(!allocate())
			return EOF;
	tbuf = new char*[++bufc_];
	for(i = 0; i < bufc_ - 1; i++)
		tbuf[i + 1] = buf_[i];
	tbuf[0] = new char[_blksize];
	delete[] buf_;
	buf_ = tbuf;
	setg(buf_[0], buf_[0] + _blksize, buf_[0] + _blksize);
	return (unsigned char)sputbackc((char)_c);
}
int dynstreambuf::do_sgetn(char* _s, int _n) {
	int _i = 0, len, _c;
	if(alloc_) {
		while(_i < _n) {
			len = (int)((bufc_ == 1 ? egptr() : pptr()) - gptr());
			len = (len > (_n - _i) ? (_n - _i) : len);
			memcpy(&(_s[_i]), gptr(), len);
			gbump(len);
			_i += len;
			if(_i >= _n || bufc_ == 1)
				break;
			if((_c = underflow()) == EOF)
				break;
			_s[_i] = (unsigned char)_c;
			_i++;
			gbump(1);
		}
	}
	return _i;
}
int dynstreambuf::do_sputn(const char* _s, int _n) {
	int _i = 0, len;
	if(!alloc_)
		if(!allocate())
			return 0;
	while(_i < _n) {
		len = (int)(epptr() - pptr());
		len = (len > (_n - _i) ? (_n - _i) : len);
		memcpy(pptr(), &(_s[_i]), len);
		pbump(len);
		_i += len;
		if(_i >= _n)
			break;
		if(overflow((unsigned char)_s[_i]) == EOF)
			break;
		_i++;
	}
	return _i;
}
