#include "../include/pipebuf.h"

const int pipestreambuf::_blksize = 1024;

pipestreambuf::pipestreambuf() : streambuf() {
	ppbase_ = ppptr_ = eppptr_ = peback_ = pgptr_ = epgptr_ = NULL;
	gbuf_ = pbuf_ = NULL;
	pput_ = put_ = 0;
	alloc_ = 0;
}
pipestreambuf::~pipestreambuf() {
	int i;
	if(alloc_) {
		for(i = 0; i <= pput_; i++) {
			if(gbuf_[i])
				delete[] gbuf_[i];
			delete[] gbuf_;
		}
		for(i = 0; i <= put_; i++) {
			if(pbuf_[i])
				delete[] pbuf_[i];
			delete[] pbuf_;
		}
	}
}
int pipestreambuf::doallocate() {
	setb((char*)1, (char*)1);
	gbuf_ = new char*[1];
	pbuf_ = new char*[1];
	gbuf_[0] = new char[_blksize];
	pbuf_[0] = new char[_blksize];
	put_ = pput_ = 0;
	setg(gbuf_[0], gbuf_[0], gbuf_[0]);
	setp(pbuf_[0], pbuf_[0] + _blksize);
	setpg(pbuf_[0], pbuf_[0], pbuf_[0] + _blksize);
	setpp(gbuf_[0], gbuf_[0] + _blksize);
	alloc_ = 1;
	return 1;
}
int pipestreambuf::sync() {
	if(!pcount()) {
		setp(pbuf_[0], pbuf_[0] + _blksize);
		setpg(pbuf_[0], pbuf_[0], pbuf_[0] + _blksize);
	}
	if(!gcount()) {
		setg(gbuf_[0], gbuf_[0], gbuf_[0]);
		setpp(gbuf_[0], gbuf_[0] + _blksize);
	}
	return (!pcount() && !gcount() ? 0 : EOF);
}
int pipestreambuf::underflow() {
	char** tbuf;
	if(pput_) {
		tbuf = new char*[pput_];
		for(int i = 1; i <= pput_; i++)
			tbuf[i - 1] = gbuf_[i];
		delete[] gbuf_[0];
		delete[] gbuf_;
		gbuf_ = tbuf;
		pput_--;
		setg(gbuf_[0], gbuf_[0], (pput_ ? gbuf_[0] + _blksize : ppptr_));
		return sgetc();
	}
	return EOF;
}
int pipestreambuf::overflow(int _c) {
	char** tbuf;
	int pdiff;
	if(!alloc_)
		return (allocate() != EOF ? sputc(_c) : EOF);
	if(pptr() >= epptr()) {
		tbuf = new char*[put_ + 2];
		for(int i = 0; i <= put_; i++)
			tbuf[i] = pbuf_[i];
		tbuf[put_ + 1] = new char[_blksize];
		delete[] pbuf_;
		pbuf_ = tbuf;
		put_++;
		setp(pbuf_[put_], pbuf_[put_] + _blksize);
	}
	return (_c != EOF ? sputc(_c) : EOF);
}
int pipestreambuf::pbackfail(int _c) {
	char** tbuf;
	if(!alloc_)
		if(!allocate())
			return EOF;
	tbuf = new char*[pput_ + 2];
	for(int i = 0; i <= pput_; i++)
		tbuf[i + 1] = gbuf_[i];
	tbuf[0] = new char[_blksize];
	delete[] gbuf_;
	gbuf_ = tbuf;
	pput_++;
	setg(gbuf_[0], gbuf_[0] + _blksize, gbuf_[0] + _blksize);
	return (unsigned char)sputbackc((char)_c);
}
int pipestreambuf::do_sgetn(char* _s, int _n) {
	int _i = 0, len, _c;
	while(alloc_ && _i < _n) {
		len = (int)((pput_ ? egptr() : ppptr_) - gptr());
		len = (len > (_n - _i) ? (_n - _i) : len);
		memcpy(&(_s[_i]), gptr(), len);
		gbump(len);
		_i += len;
		if(_i >= _n || !pput_)
			break;
		if((_c = underflow()) == EOF)
			break;
		_s[_i] = (unsigned char)_c;
		_i++;
		gbump(1);
	}
	return _i;
}
int pipestreambuf::do_sputn(const char* _s, int _n) {
	int _i = 0, len;
	if(!alloc_)
		if(!allocate())
			return 0;
	while(_i < _n) {
		len = (int)(epptr() - pptr());
		len = (len > (_n - _i) ? (_n - _i) : len);
		memcpy(pptr(), &(_s[_i]), len);
		pbump(len);
		_i += len;
		if(_i >= _n)
			break;
		if(overflow((unsigned char)_s[_i]) == EOF)
			break;
		_i++;
	}
	return _i;
}
int pipestreambuf::punderflow() {
	char** tbuf;
	if(put_) {
		tbuf = new char*[put_];
		for(int i = 1; i <= put_; i++)
			tbuf[i - 1] = pbuf_[i];
		delete[] pbuf_[0];
		delete[] pbuf_;
		pbuf_ = tbuf;
		put_--;
		setpg(pbuf_[0], pbuf_[0], pbuf_[0] + _blksize);
		return pgetc();
	}
	return EOF;
}
int pipestreambuf::poverflow(int _c) {
	char** tbuf;
	int pdiff;
	if(!alloc_)
		return (allocate() != EOF ? pputc(_c) : EOF);
	if(ppptr_ >= eppptr_) {
		tbuf = new char*[pput_ + 2];
		for(int i = 0; i <= pput_; i++)
			tbuf[i] = gbuf_[i];
		tbuf[pput_ + 1] = new char[_blksize];
		delete[] gbuf_;
		gbuf_ = tbuf;
		pput_++;
		setpp(gbuf_[pput_], gbuf_[pput_] + _blksize);
	}
	return (_c != EOF ? pputc(_c) : EOF);
}
int pipestreambuf::ppbackfail(int _c) {
	char** tbuf;
	if(!alloc_)
		if(!allocate())
			return EOF;
	tbuf = new char*[put_ + 2];
	for(int i = 0; i <= put_; i++)
		tbuf[i + 1] = pbuf_[i];
	tbuf[0] = new char[_blksize];
	delete[] pbuf_;
	pbuf_ = tbuf;
	put_++;
	setg(pbuf_[0], pbuf_[0] + _blksize, pbuf_[0] + _blksize);
	return pputbackc((char)_c);
}
int pipestreambuf::do_pgetn(char* _s, int _n) {
	int _i = 0, len, _c;
	while(alloc_ && _i < _n) {
		len = (int)((put_ ? epgptr_ : pptr()) - pgptr_);
		len = (len > (_n - _i) ? (_n - _i) : len);
		memcpy(&(_s[_i]), pgptr_, len);
		pgbump(len);
		_i += len;
		if(_i >= _n || !put_)
			break;
		if((_c = punderflow()) == EOF)
			break;
		_s[_i] = _c;
		_i++;
		pgbump(1);
	}
	return _i;
}
int pipestreambuf::do_pputn(const char* _s, int _n) {
	int _i = 0, len;
	if(!alloc_)
		if(!allocate())
			return 0;
	while(_i < _n) {
		len = (int)(eppptr_ - ppptr_);
		len = (len > (_n - _i) ? (_n - _i) : len);
		memcpy(ppptr_, &(_s[_i]), len);
		ppbump(len);
		_i += len;
		if(_i >= _n)
			break;
		if(poverflow((unsigned char)_s[_i]) == EOF)
			break;
		_i++;
	}
	setg(eback(), gptr(), (pput_ ? gbuf_[0] + _blksize : ppptr_));
	return _i;
}
int pipestreambuf::do_pnextc() {
	char** tbuf;
	if(put_) {
		tbuf = new char*[put_];
		for(int i = 1; i <= put_; i++)
			tbuf[i - 1] = pbuf_[i];
		delete[] pbuf_;
		pbuf_ = tbuf;
		put_--;
		setpg(pbuf_[0], pbuf_[0], (put_ ? pbuf_[0] + _blksize : pptr()));
		return pgetc();
	}
	return EOF;
}
